/*******************************************************************************
 * Copyright 2011. Kronos Incorporated. All right reserved.
 ******************************************************************************/
package com.kronos.wfc.platform.security.business.authentication.filters;

import com.kronos.wfc.platform.logging.framework.Log;
import com.kronos.wfc.platform.logging.framework.LogContext;
import com.kronos.wfc.platform.properties.framework.KronosProperties;

import javax.servlet.*;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Abstract class that incorporates the {@link javax.servlet.Filter} interface
 * and adds logic to define a log context.
 * <p/>
 *
 * @author Douglas Ross
 * @author John Richards
 */
public abstract class WFPFilter implements Filter {
  ///////////////////////////////////////////////////////////////////
  //
  // Class statics
  //
  ///////////////////////////////////////////////////////////////////

  /**
   * Defines a constant value for a comma.
   */
  private static final String DEFAULT_EXCEPTION_LIST = ".*jar$, .*gif$, .*jpg$, .*jpeg$, .*[.]css$, .*js$, .*class$, .*htm$, .*html$";

  /**
   * Defines the property key used to set the default exception list.
   */
  private static final String DEFAULT_EXCEPTION_LIST_PROPERTY = "com.kronos.wfc.platform.security.business.authentication.filters.DefaultExceptionList";

  /**
   * Defines a constant value for a comma.
   */
  private static final String COMMA = ",";

  /**
   * Defines the value for {@link java.lang.String} operations in which sub strings are not found.
   */
  protected static final int STRING_NOT_FOUND = -1;

  /**
   * Track whether the exceptions have been initialized.
   */
  private boolean exceptionsInitialized = false;

  ////////////////////////////////////////////////////////////////////
  //
  // This section defines the class fields
  //
  ////////////////////////////////////////////////////////////////////

  /**
   * A collection of URI patterns that do not need to be processed by
   * the filter.
   */
  private List exceptionList = new ArrayList();

  /**
   * The Kronos log context for all filters.
   * Sub-classes may set their own log context to narrow the scope
   * of the filter for the log.
   */
  private LogContext logContext = new LogContext(Log.WFP, "FILTER");

  /**
   * A handle to the ServletContext object
   */
  protected ServletContext servletContext = null;

  ///////////////////////////////////////////////////////////////////
  //
  // Class constructors
  //
  ///////////////////////////////////////////////////////////////////

  /**
   * Default Constructor.
   * Constructs a filter.
   */
  public WFPFilter() {
    super();

    // Only process URIs that are private. Images, style sheets, JSP, JavaScript need not
    // be validated for caching.
    setExceptionList(DEFAULT_EXCEPTION_LIST_PROPERTY, DEFAULT_EXCEPTION_LIST);
  }

  ///////////////////////////////////////////////////////////////////
  //
  // The field accessor methods
  //
  ///////////////////////////////////////////////////////////////////

  /**
   * Return the list of URI exceptions.
   * Return the collection of URI patterns that do not need to be processed by
   * the filter.
   *
   * @return collection of URI patterns that need not be processed
   */
  public List getExceptionList() {
    return exceptionList;
  }

  /**
   * Set the list of URI exceptions.
   * Update the collection of URI patterns that do not need to be processed by
   * the filter.
   *
   * @param exceptionList collection of URI patterns that need not be processed
   */
  protected void setExceptionList(List exceptionList) {
    this.exceptionList = exceptionList;
  }

  /**
   * Set the list of URI exceptions.
   * Update the collection of URI patterns that do not need to be processed by
   * the filter.
   * <p/>
   * If the property is not found or the property contains no data, then the current
   * exclusion list will remain intact.
   *
   * @param propertyName Kronos property used to contain the collection of URIs that are not to be processed
   */
  protected void setExceptionList(String propertyName) {
    setExceptionList(propertyName, null);
  }

  /**
   * Set the list of URI exceptions.
   * Update the collection of URI patterns that do not need to be processed by
   * the filter.
   * <p/>
   * If the property is not found or the property contains no data, then the current
   * exclusion list will remain intact.
   *
   * @param propertyName Kronos property used to contain the collection of URIs that are not to be processed
   * @param defaultValues A common separated list that contains the default values if the poperty cannot be found.
   * It may be null.
   */
  protected void setExceptionList(String propertyName, String defaultValues) {
    // Set the default to null
    List list = null;

    // Get the exceptions from the property
    String exceptions = KronosProperties.getProperty(propertyName, defaultValues);
    if (exceptions != null) {
      // Create the list of exceptions
      list = buildExceptionList(exceptions);
    }

    // Always create an empty list
    if (list == null || getExceptionList() == null) {
      list = new ArrayList();
    }
    setExceptionList(list);
  }

  /**
   * Return the context for logging.
   *
   * @return the log context
   */
  public LogContext getLogContext() {
    return logContext;
  }

  /**
   * Set the logging context name for this instance of the filter.
   *
   * @param name Context name used for logging.
   * @deprecated Override getLogContext() instead.
   */
  protected void setLogContext(String name) {
  }

  /**
   * Return the property key used for getting exceptions for the filter.
   *
   * @return property name
   */
  protected String getExceptionProperty() {
    return DEFAULT_EXCEPTION_LIST_PROPERTY;
  }

  ///////////////////////////////////////////////////////////////////
  //
  // This section implements the {@link javax.servlet.Filter} methods
  //
  ///////////////////////////////////////////////////////////////////

  /**
   * Process the configuration for the filter.
   * Review and process any parameters associated with the filter.
   * <p/>
   * TODO: process any config params loaded from app-config.xml
   *
   * @param config Filter configuration
   */
  public void init(FilterConfig config) throws ServletException {
    servletContext = config.getServletContext();
  }

  /**
   * Process the {@link javax.servlet.Filter}
   * Method is abstract in order to force sub-classes to provide their own operations.
   * <p/>
   * <pre>
   * 		if (isFilterOk(req, res)) {
   * 			chain.doFilter(req, res);
   *     } else {
   * 			WFPSystem.forwardRequest("my-bad-filter.jsp", req, res);
   *     }
   * </pre>
   * <p/>
   *
   * @param req A request from an http session.
   * @param res The response for the http request.
   * @param chain The chain of filters.
   * @throws java.io.IOException            Included as part of the Filter interface
   * @throws javax.servlet.ServletException Included as part of the Filter interface
   */
  abstract public void doFilter(
      ServletRequest req,
      ServletResponse res,
      FilterChain chain)
      throws IOException, ServletException;

  /**
   * Reset any instance variables, so this instance can be re-used.
   * Method is abstract in order to force sub-classes to provide their own operations.
   */
  abstract public void destroy();

  ///////////////////////////////////////////////////////////////////
  //
  // This section contains the private and protect methods
  //
  ///////////////////////////////////////////////////////////////////

  /**
   * Determine if the URI needs to be processed by the filter.
   * Use this method to insure that a filter only needs to process a specific
   * set of URI patterns.
   *
   * @param uri the URI from the request
   * @return true when the URI needs to be processed; false otherwise. Use
   */
  protected boolean isException(String uri) {
    // Assume the URI is not on the exception list
    boolean isException = false;

    // Update the list of exceptions
    updateExceptions();

    List exceptions = getExceptionList();
    if (exceptions != null) {
      for (Iterator iterator = exceptions.iterator(); iterator.hasNext(); ) {
        String exception = (String) iterator.next();
        if(exception == null) {
        	continue;
        }
        Pattern pattern = Pattern.compile(exception);
        Matcher matcher = pattern.matcher(uri);
        if (matcher.matches()) {
          isException = true;
          break;
        }
      }
    }

    // Return the status
    return isException;


  }

  /**
   * Returns a handle to the ServletContext object
   *
   * @return ServletContext the ServletContext object
   */
  protected ServletContext getServletContext() {
    return servletContext;
  }

  /**
   * Adds data to the list of URI exceptions.
   * Update the collection of URI patterns that do not need to be processed by
   * the filter by adding data to the associated list.
   *
   * @param exceptions collection of URI patterns that need not be processed
   */
  protected void addToExceptionList(List exceptions) {
    exceptionList.addAll(exceptions);
  }

  /**
   * Removes data from the list of URI exceptions.
   * Update the collection of URI patterns that do not need to be processed by
   * the filter by removing data from teh asscoaited list.
   *
   * @param exceptions collection of URI patterns that need not be processed
   */
  protected void removeFromExceptionList(List exceptions) {
    exceptionList.removeAll(exceptions);
  }

  /**
   * Creates a List of exceptions for the filter.
   * Takes the comma separated list an creates a List of strings that
   * define the exceptions for the filter.
   *
   * @param exceptions Comma separated list of <b>patterns</b> to not include
   * @return a list of exceptions for the filter
   */
  private List buildExceptionList(String exceptions) {
    // Create the container for the exceptions
    List list = new ArrayList();

    // Parse the string and add the elements to the list
    StringTokenizer tokenizer = new StringTokenizer(exceptions, COMMA);
    while (tokenizer.hasMoreElements()) {
      // The exceptions are all file extensions, and need to be converted
      // into regular expressions.
      String exception = (String) tokenizer.nextElement();
      list.add(exception.trim());
    }
    return list;
  }

  /**
   * Update the list of exceptions.
   * Read the exceptions list from a property and update the list with changes.
   */
  synchronized protected void updateExceptions() {
    if (!exceptionsInitialized) {
      // Get the exceptions from the property
      List list = null;
      String key = getExceptionProperty();
      String exceptions = KronosProperties.getProperty(key, null);
      if (Log.priorityEnabled(getLogContext(), Log.DEBUG)) {
        Log.log(getLogContext(),
                Log.DEBUG,
                "Exceptions for filter",
                key,
                exceptions);
      }
      if (exceptions != null) {
        // Create the list of exceptions
        list = buildExceptionList(exceptions);
        if (list != null && list.size() > 0) {
          addToExceptionList(list);
        }
      }
      exceptionsInitialized = true;
    }
  }


} // End WFPFilter
